/*Copyright (c) 2012 Moussa Dembélé (www.mousman.com)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/
package com.mousman.as3.agal.parser.objs 
{
	import com.greensock.TweenLite;
	import com.mousman.as3.agal.parser.dto.MBuffersWriterDTO;
	import com.mousman.as3.agal.parser.verticeswriter.WriterSet;
	import com.mousman.as3.agal.parser.verticeswriter.WriterStore;
	import com.mousman.as3.agal.parser.verticeswriter.WriterType;
	import com.mousman.as3.structure.garbage.IDestroyable;
	import com.mousman.as3.utils.color.ColorUtil;
	import flash.display3D.Context3D;
	import flash.display3D.IndexBuffer3D;
	import flash.display3D.VertexBuffer3D;
	import flash.events.Event;
	import flash.events.EventDispatcher;
	import flash.utils.ByteArray;
	import org.osflash.signals.Signal;
	/**
	 * creates a set of vertex buffers and index buffers based on a .obj and .mtl file
	 * and exports them in a binary file
	 * @author Moussa Dembélé (www.mousman.com) based on Christer Kaitila and Nathan Ostgard work
	 * 
	 */
	public class MObjExportParser extends EventDispatcher implements IDestroyable
	{
		// constants used in parsing OBJ data
		private const LINE_FEED:String = String.fromCharCode(10);
		private const SPACE:String = String.fromCharCode(32);
		private const SLASH:String = "/";
		private const VERTEX:String = "v";
		private const NORMAL:String = "vn";
		private const UV:String = "vt";
		private const INDEX_DATA:String = "f";	
		private const MTL_FILE:String = "mtllib";	
		private const USE_MTL:String = "usemtl";	
		private const OBJECT_NAME:String = "o";	
		private const GROUP_NAME:String = "g";	
		private const SMOOTH_SHADING:String = "s";
		
		private var _mtlFile:String;		
		// older versions of 3dsmax use an invalid vertex order:
		private var _vertexDataIsZxy:Boolean = false;
		
		private var _scale:Number;
		private var _vertices:Vector.<Number>;
		private var _normals:Vector.<Number>;
		private var _uvs:Vector.<Number>;

		
		// the parser for the mtl datas
		private var _mtlParser:MMtlParser;
		// contains all the mtl datas parsed for the mtl file
		private var _mtlSets:Vector.<MMtlSet>;
		
		private var _currentObject:MObject3D;
		private var _currentGroup:MPolygoneGroup3D;
		// contains all the polygons group for the current object
		private var _mGroups:Vector.<MPolygoneGroup3D>;
		
		private var _currentWriterSet:WriterSet;
		private var _defaultWriterSet:WriterSet;
		private var _defaultMtlSet:MMtlSet;
		
		// contains all the datas for the mesh
		private var _mesh:MObjMesh;
		
		// contains colors that we can force in the construction of the mesh
		private var _writerParams:MBuffersWriterDTO;
		// contains all the shaders we want to use
		private var _writerStore:WriterStore;
		// some exporters mirror the UV texture coordinates
		private var _mirrorUV:Boolean = true;
		
		private var _lineSignal:Signal;
		private var _signal:Signal;
		
		private var _lines:Array;
		private var _currentLineIndex:int;
		private const MAX_POLYGONES_PER_GROUP:int = 21248;
		private var _currentPolygoneNumber:int = 0;		
		private var _totalPolygoneNumber:int = 0;		
		
		/**
		 * @private
		 */		
		private var _context3D:Context3D;
		
		/**
		* Constructor. Creates a new MObjParser instance.
		*
		* <p>After you create the MObjParser instance, call the parse method to create the mesh datas. And then the export method</p>
		* 
		* @param writerStore : WriterStore.<p>Used to resize the object</p>
		* @param dataIsZxy : Boolean.<p>Older versions of 3dsmax use an invalid vertex order</p>
		* @param mirrorUV : Boolean.<p>Some exporters mirror the UV texture coordinates</p>
		*/		
		public function MObjExportParser(writerStore:WriterStore, dataIsZxy:Boolean = false, mirrorUV:Boolean = true) {
			// older versions of 3dsmax use an invalid vertex order:
			_vertexDataIsZxy = dataIsZxy;
			// some exporters mirror the UV texture coordinates
			_mirrorUV = mirrorUV;
			_writerStore = writerStore;
		}
		//----------------------------------------------------------------------------------------------
		/**
		* Parse the obj and mtl file and create the mesh datas
		*
		* 
		* @param objfile : String.<p>Contains the obj file datas</p>
		* @param mtlfile : String.<p>Contains the mtl file datas</p>
		* @param scale : Number.<p>Used to resize the 3D model</p>
		* @param writerParams : used to force color on the mesh.<p></p>
		* 
		* @return the datas of the mesh
		*/			
		//----------------------------------------------------------------------------------------------		
		public function parse(objfile:String, mtlfile:String = null, scale:Number=1, writerParams:MBuffersWriterDTO = null): MObjMesh {
			_mesh = new MObjMesh();
			_scale = scale;
			TweenLite.delayedCall(1,startParsing,[objfile,mtlfile,writerParams]);
			
			return _mesh;
		}
		//----------------------------------------------------------------------------------------------
		//----------------------------------------------------------------------------------------------
		private function startParsing(objfile:String, mtlfile:String = null, writerParams:MBuffersWriterDTO = null):void 
		{
			if(writerParams == null) 
			_writerParams = new MBuffersWriterDTO();
			// read and explode the mtl datas
			if(mtlfile)
				parseMtl(mtlfile);
			_defaultMtlSet = new MMtlSet("default");
			_defaultMtlSet.ambiantColor = 0xcccccc;
			_defaultMtlSet.diffuseColor = 0xcccccc;
			_defaultMtlSet.dissolved= 1;
			_defaultMtlSet.emissiveColor = 0x0;
			_defaultMtlSet.specularCoef = 100;
			_defaultMtlSet.specularColor= 0xffffff;
			// read and explode the obj datas
			parseLines(objfile);
		}
		//----------------------------------------------------------------------------------------------
		// parse the mtl file and create a set of MMtlSet
		//----------------------------------------------------------------------------------------------
		private function parseMtl(mtlfile:String):void {
			_mtlParser = new MMtlParser(mtlfile);
			_mtlParser.parse(_writerParams);
			_mtlSets = _mtlParser.mtlSetVector;
		}	
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function parseLines(objfile:String):void {
			_lines = objfile.split(LINE_FEED);
			
			var line:String;
			var loop:uint = _lines.length;		
			var id:String;
			var writerSet:WriterSet;	
			
			_vertices = new Vector.<Number>();
			_normals = new Vector.<Number>();		
			_uvs = new Vector.<Number>();
			
			// initialisation of the writers, passing them vertices, normals and _uv references
			for each(writerSet in _writerStore.writerSetArray) {
				writerSet.writer.init(_vertices, _normals,_uvs);
			}
			
			if (_lineSignal) _lineSignal.remove(parseBlock);
			_lineSignal = new Signal();
			_lineSignal.add(parseBlock);
			_currentLineIndex = 0;
			parseBlock();
			
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function parseBlock():void 
		{
			log("parseBlock");
			nextBlock();
		}	
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function nextBlock():void 
		{
			var line:String;
			var out:Boolean = false;
		
			var loop:int = 500;
			var max:int = _lines.length;
			while(!out && _currentLineIndex <= max) {
				// last line
				if (_currentLineIndex == max) {
					// complete last polygone group
					log("complete");
					_mesh.nbVertices = _totalPolygoneNumber*3;
					completePolygonGroup();
					dispatchEvent(new Event(Event.COMPLETE));
					return;
				}
				line = _lines[_currentLineIndex];
				line.replace(/^\s*|\s*$/g, '');
				parseLine(line);
				_currentLineIndex ++;
				if (_currentPolygoneNumber == MAX_POLYGONES_PER_GROUP) {
					// we take the mtlSet of the current group
					var mtlSet:MMtlSet = _currentGroup.mtlSet;
					completePolygonGroup();
					// we create a group for the next bloc of polygones and assign the mtl data of the current group
					_currentGroup = new MPolygoneGroup3D(_currentLineIndex.toString());
					_currentGroup.mtlSet = mtlSet;
					if (mtlSet.isTexture) {
						_currentWriterSet = _writerStore.getWriterById(WriterType.DEFAULT_TEXTURE_WRITER);
					}
					else {
						_currentWriterSet = _writerStore.getWriterById(WriterType.DEFAULT_WRITER);
					}					
					// there is no object 
					if (!_currentObject) 
						setNewObject(_currentLineIndex.toString());
					_mGroups.push(_currentGroup);					
					_currentPolygoneNumber = 0;
					out = true;
				}
			}
			
			if (out) {
				callNext();
			}
		}		
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		public function callNext():void {
			_lineSignal.dispatch();
		}		
		//----------------------------------------------------------------------------------------------	
		//----------------------------------------------------------------------------------------------		
		private function parseLine(line:String):void {
			var name:String;
			// Split line into words.
			var words:Array = line.split(SPACE);
			// Blank line or comment, ignore it
			if (words.length ==0 || line === '' || line.charAt(0) === '#') {
				return;
			}
			// Prepare the data of the line.
			var data:Array = words.slice(1);
			
			// Check first word and delegate remainder to proper parser.
			while ((data[0] == '') || (data[0] == ' ')) data.splice(0, 1);
			var firstWord:String = words[0];
			name = data[0] as String;
			name = name.substr(0, name.length - 1);
			
			switch (firstWord)
			{
				case MTL_FILE:
					log("mtl file : " + data[0]);
					_mtlFile = data[0];
				break;
				case OBJECT_NAME:
					log("obj : " + name);
					// let's finish the creation of the last polygon group for the precedent object (if exist)
					if (_currentGroup) {
						completePolygonGroup();
					}
					setNewObject(name);
					// if we have a writer defined for this object
					if (_writerStore.getWriterById(name)) {
						_currentWriterSet = _writerStore.getWriterById(name);
					}
				break;
				case VERTEX:
					parseVertex(data);
				break;
				case NORMAL:
					parseNormal(data);
				break;	
				case UV:
					parseUv(data);
				break;				
			case GROUP_NAME:
					log("   group : " + name);
					// let's finish the creation of the precedent polygon group
					if (_currentGroup) {
						completePolygonGroup();
					}
					_currentGroup = new MPolygoneGroup3D(name);
					// a group exist and is not part of an object 
					if (!_currentObject) setNewObject(name);
					_mGroups.push(_currentGroup);
					// we look for a writer associated with the group 
					if(_writerStore.getWriterById(name)){
						_currentWriterSet = _writerStore.getWriterById(name); 
					}
					// else we use the one define for the object if it exists 
					else if(_writerStore.getWriterById(_currentObject.id)){
						_currentWriterSet = _writerStore.getWriterById(_currentObject.id); 
					} 
					else
						_currentWriterSet = _writerStore.getWriterById(WriterType.DEFAULT_WRITER);
				break;				
				case USE_MTL:
					log("   mtl : " + data[0]);
					var  mtlID:String = data[0];
					var found:Boolean = false;
					// we look for the mtl set corresponding 
					for each(var mtlSet:MMtlSet in _mtlSets) {
						if (mtlSet.id == mtlID) {
							_currentGroup.mtlSet = mtlSet;
							found = true;
						}
					}
					if(!found) 
						_currentGroup.mtlSet = _defaultMtlSet;
					
					// no WriterSet has been defined so we use the default ones
					if (!_currentWriterSet) {
						if (_currentGroup.mtlSet.isTexture) {
							_currentWriterSet = _writerStore.getWriterById(WriterType.DEFAULT_TEXTURE_WRITER);
						}
						else {
							_currentWriterSet = _writerStore.getWriterById(WriterType.DEFAULT_WRITER);
						}
					}
					// set the diffuse color as vertex color for the group
					var color:Object = ColorUtil.hexArgb(_currentGroup.mtlSet.diffuseColor);
					var alpha:int = _currentGroup.mtlSet.dissolved * 255;
					var tcolor:uint = ColorUtil.argbHex(alpha, color.bleu, color.vert, color.rouge);
					_currentWriterSet.writer.verticesColor = tcolor;
				break;		
				case SMOOTH_SHADING:
					_currentGroup.smoothShading = Number(data[0]);
				break;	
				case INDEX_DATA:
					parseFaces(data);
				break;
			}
		}
		//----------------------------------------------------------------------------------------------	
		// create a new object
		//----------------------------------------------------------------------------------------------		
		private function setNewObject(data:String):void {
			_currentObject = new MObject3D(data);
			_mesh.add(_currentObject);
			_mGroups = new Vector.<MPolygoneGroup3D>();
			_currentObject.polygonsGroups = _mGroups;		
		}
		//----------------------------------------------------------------------------------------------	
		//----------------------------------------------------------------------------------------------		
		private function completePolygonGroup():void {
			_currentWriterSet.writer.complete();
			createBuffers();
			_currentGroup = null;
			_currentWriterSet.writer.reset();
			_currentWriterSet = null;
		}
		//----------------------------------------------------------------------------------------------
		// explode the face into triangles 
		//----------------------------------------------------------------------------------------------		
		private function parseFaces(data:Array):void{
			var nbVertices:int = data.length;
			var triplet:Vector.<String> = new Vector.<String>(3, true);
			// Process elements turning faces into triangle fan.
			for (var i:int = 1; i < nbVertices-1; ++i) {
				triplet[0] = data[i];
				triplet[1] = data[0];
				triplet[2] = data[i + 1];
				parseTriplet(triplet);
				_currentPolygoneNumber ++;
				_totalPolygoneNumber ++;				
			}
		}
		//----------------------------------------------------------------------------------------------	
		// for each triangle pass datas to the writer wich will add them in a byteArray
		//----------------------------------------------------------------------------------------------		
		private function parseTriplet(data:Vector.<String>):void {
			var triplet:String;
			var indexTriplet:Vector.<uint> = new Vector.<uint>(3,true);
			var normalTriplet:Vector.<uint> = new Vector.<uint>(3,true);
			var uvTriplet:Vector.<uint> = new Vector.<uint>(3,true);
			var subdata:Array;
			var vertexIndex:uint;
			var normalIndex:uint;
			var uvIndex:uint;
			var index:uint;
			var loop:int = 3;
			var i:int;
			
			// loop through each element and grab values stored earlier
			// elements come as vertexIndex/normalIndex
			for(i = 0; i < loop; ++i){
				triplet = data[i];
				subdata = triplet.split(SLASH);
				vertexIndex = uint(subdata[0]) - 1;
				uvIndex = uint(subdata[1]) - 1;
				normalIndex = uint(subdata[2]) - 1;
				
				// sanity check
				if(vertexIndex < 0) vertexIndex = 0;
				if (normalIndex < 0) normalIndex = 0;
				if (uvIndex < 0) uvIndex = 0;
				
				indexTriplet[i] = vertexIndex;
				normalTriplet[i] = normalIndex;
				uvTriplet[i] = uvIndex;
			}
			
			_currentWriterSet.writer.addTriangle(indexTriplet, normalTriplet, uvTriplet);
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function parseVertex(data:Array):void {
			if (_vertexDataIsZxy){
				_vertices.push(Number(data[1])*_scale);
				_vertices.push(Number(data[2])*_scale);
				_vertices.push(Number(data[0])*_scale);
			}
			else { // normal operation: x,y,z
				_vertices.push(Number(data[0])*_scale);
				_vertices.push(Number(data[1])*_scale);
				_vertices.push(Number(data[2])*_scale);
			}
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function parseNormal(data:Array):void{
			var loop:uint = data.length;
			if (loop > 3) {
				log("MObjParser more than 3 normales !!!");
				loop = 3;
			}
			for (var i:uint = 0; i < loop; ++i){
				var element:String = data[i];
				if (element != null) // handle 3dsmax extra spaces
				_normals.push(Number(element));
			}
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function parseUv(data:Array):void{
				var element:String = data[0];
				if(_mirrorUV)
				//OBJ has a flipped U axis for texture coordinates, so it needs to get flipped back to normal:
					_uvs.push(1 - Number(element));
				else
					_uvs.push(Number(element));
				
				element = data[1];
				if(_mirrorUV)
				//OBJ has a flipped V axis for texture coordinates, so it needs to get flipped back to normal:
					_uvs.push(1 - Number(element));
				else
					_uvs.push(Number(element));
		}
		//----------------------------------------------------------------------------------------------		
		// create the VertexBuffer3D, IndexBuffer3D and corresponding bytesArrays for the current group
		//----------------------------------------------------------------------------------------------		
		private function createBuffers():void {
			var vertexByteArray:ByteArray;
			var rawIndexBuffer:Vector.<uint>;
			var pGroupBuffers:MPGBuffersExport;
			var nbVertices:int;
			var i:int;
			var dataPerVertex:int = _currentWriterSet.dataPerVertex;
			var ba:ByteArray;
			var vertexBuffer:VertexBuffer3D;
			var indexBuffer:IndexBuffer3D;			

			nbVertices = _currentWriterSet.writer.nbVertices;
			_currentGroup.nbVertices = nbVertices;
			vertexByteArray = _currentWriterSet.writer.bytesArray;
			
			var next:int = 0;
			var rest:int = 0;
			var uploaded:int = 0;

			rest = nbVertices - uploaded;
			while (rest) {
				if ( rest > 65535) rest = 65535;
				pGroupBuffers = new MPGBuffersExport(_currentGroup.id+"_"+i);
				 //vertices
				pGroupBuffers.dataPerVertex = dataPerVertex;
				pGroupBuffers.nbVertices = rest;
				ba = new ByteArray();
				ba.writeBytes(vertexByteArray, uploaded * dataPerVertex * 4, rest * dataPerVertex * 4);
				pGroupBuffers.vertexBufferBA = ba;
				
				 //index
				rawIndexBuffer = _currentWriterSet.writer.indices.slice(uploaded, uploaded + rest);
				pGroupBuffers.indexBufferV = rawIndexBuffer;
				pGroupBuffers.nbIndices = rest;
				
				if (_context3D) {
					 //vertices
					vertexBuffer = _context3D.createVertexBuffer(rest, dataPerVertex);
					//vertexBuffer.uploadFromByteArray(ba,0,0,rest);
					vertexBuffer.uploadFromByteArray(vertexByteArray, uploaded * dataPerVertex * 4, 0, rest);
					 //index
					indexBuffer = _context3D.createIndexBuffer(rest);
					indexBuffer.uploadFromVector(rawIndexBuffer, 0, rest);
					
					pGroupBuffers.vertexBuffer = vertexBuffer;
					pGroupBuffers.indexBuffer = indexBuffer;					
				}
				
				_currentGroup.addPGBExport(pGroupBuffers);
				
				uploaded += rest;
				rest = nbVertices - uploaded;
				i++;
			}
		}
		//----------------------------------------------------------------------------------------------
		/**
		* create a binary file containing the obj and mtl datas
		* @return object containing mesh datas in a binary format
		*/			
		//----------------------------------------------------------------------------------------------
		public function export():ByteArray {
			var object3Dset:MObject3D;
			var mPolygonsGroup:MPolygoneGroup3D;
			var totalVertices:int;
			var startObjPos:uint;
			var endObjPos:uint;
			var size:uint;
			
			
			var exportBA:ByteArray = new ByteArray();
			createHeader(exportBA);
			
			exportBA.writeUnsignedInt(_mesh.mObjects.length);			
			for each(object3Dset in _mesh.mObjects) {
				log("--------  object " + object3Dset.id);
				// writting a object3Dset
				startObjPos = exportBA.position;
				exportBA.writeUnsignedInt(0);
				exportBA.writeUTF(object3Dset.id);
				exportBA.writeUnsignedInt(object3Dset.polygonsGroups.length);	
				
				for each(mPolygonsGroup in object3Dset.polygonsGroups) {
					writePGroup(exportBA,mPolygonsGroup);
				}
				
				endObjPos = exportBA.position;
				size = endObjPos - startObjPos;
				exportBA.position = startObjPos;
				exportBA.writeUnsignedInt(size);
				exportBA.position = endObjPos;
			}
			log("export complete");
			return exportBA;
		}
		//----------------------------------------------------------------------------------------------
		//----------------------------------------------------------------------------------------------
		private function writePGroup(exportBA:ByteArray,mPolygonsGroup:MPolygoneGroup3D):void {
			var startPGroupPos:uint;
			var endPGroupPos:uint;
			var size:uint;
			var startMtlPos:uint;
			var endMtlPos:uint;
			var mtlSet:MMtlSet;
			var polygonBufferSet:MPGBuffersExport;
			
			//save position of the mPolygonsGroup size
			startPGroupPos = exportBA.position;
			exportBA.writeUnsignedInt(0);	
			exportBA.writeUTF(mPolygonsGroup.id);
			exportBA.writeUnsignedInt(mPolygonsGroup.nbVertices);				
			exportBA.writeDouble(mPolygonsGroup.smoothShading);	
			
			// write mtl set
			mtlSet = mPolygonsGroup.mtlSet;
			if (!mtlSet) {
				mtlSet = new MMtlSet("MDefault");
				mtlSet.ambiantColor = 0xcccccc;
				mtlSet.diffuseColor = 0xcccccc;
				mtlSet.dissolved= 1;
				mtlSet.emissiveColor = 0x0;
				mtlSet.specularCoef = 100;
				mtlSet.specularColor= 0xffffff;
			}
			startMtlPos = exportBA.position;
			exportBA.writeUnsignedInt(0);
			exportBA.writeUTF(mPolygonsGroup.id);
			exportBA.writeUnsignedInt(mtlSet.diffuseColor);
			exportBA.writeUnsignedInt(mtlSet.ambiantColor);
			exportBA.writeUnsignedInt(mtlSet.specularColor);
			exportBA.writeDouble(mtlSet.specularCoef);	
			exportBA.writeUnsignedInt(mtlSet.emissiveColor);
			exportBA.writeDouble(mtlSet.dissolved);
			exportBA.writeDouble(mtlSet.illumination);
			exportBA.writeUTF(mtlSet.ambiantTexture);
			exportBA.writeUTF(mtlSet.diffuseTexture);
			exportBA.writeBoolean(mtlSet.isTexture);
			endMtlPos = exportBA.position;
			size = endMtlPos - startMtlPos;
			exportBA.position = startMtlPos;
			exportBA.writeUnsignedInt(size);
			exportBA.position = endMtlPos;
			
			// write polygonGroup buffers
			exportBA.writeUnsignedInt(mPolygonsGroup.pGBExportSet.length);		
			for each(polygonBufferSet in mPolygonsGroup.pGBExportSet) {
				writeBufferSet(exportBA, polygonBufferSet);
			}
			
			endPGroupPos = exportBA.position;
			size = endPGroupPos - startPGroupPos;
			exportBA.position = startPGroupPos;
			exportBA.writeUnsignedInt(size);
			exportBA.position = endPGroupPos;
		}
		//----------------------------------------------------------------------------------------------
		//----------------------------------------------------------------------------------------------
		private function writeBufferSet(exportBA:ByteArray, polygonBufferSet:MPGBuffersExport):void 
		{
			var startBSPos:uint;
			var endBSPos:uint;		
			var size:uint;
			
			startBSPos = exportBA.position;
			exportBA.writeUnsignedInt(0);
			exportBA.writeUTF(polygonBufferSet.id);
			exportBA.writeInt(polygonBufferSet.dataPerVertex);
			exportBA.writeInt(polygonBufferSet.nbVertices);
			exportBA.writeBytes(polygonBufferSet.vertexBufferBA);
			exportBA.writeInt(polygonBufferSet.nbIndices);
			for each(var indice:uint in polygonBufferSet.indexBufferV) {
				exportBA.writeUnsignedInt(indice);
			}
			endBSPos = exportBA.position;
			size = endBSPos - startBSPos;
			exportBA.position = startBSPos;
			exportBA.writeUnsignedInt(size);
			exportBA.position = endBSPos;
		}
		//----------------------------------------------------------------------------------------------
		//----------------------------------------------------------------------------------------------
		private function createHeader(exportBA:ByteArray):void 
		{
			exportBA.writeUTFBytes("MDM");
			var empty:String = "";
			for (var i:int = 0; i < 99; i++) empty += "m";
			exportBA.writeUTFBytes(empty);
		}		
		//----------------------------------------------------------------------------------------------
		//----------------------------------------------------------------------------------------------
		private function log(txt:String):void 
		{
			trace(txt);
		}		
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function clean():void {
		}
		//----------------------------------------------------------------------------------------------	
		/**
		* delete datas
		*/				
		//----------------------------------------------------------------------------------------------	
		public function destroy():void {
			_mtlFile = null; 
			_vertices.length = 0;
			_vertices = null;
			_normals.length = 0;
			_normals = null;
			_uvs.length = 0;
			_uvs = null;
			_mtlParser.destroy();
			_mtlParser = null;
			_mtlSets.length = 0;
			_mtlSets = null;
			_currentObject = null;
			_currentGroup = null;
			_mGroups.length = 0;
			_mGroups = null;
			_currentWriterSet = null;
			_defaultWriterSet = null;
			_mesh.destroy();
			_mesh = null;
			_writerParams = null;
			_writerStore.destroy();
			_writerStore = null;
			_lineSignal.removeAll();
			_lineSignal= null;
		}
		//----------------------------------------------------------------------------------------------	
		/**
		* context3D used for creating and uploading vertex3Ds and index3Ds
		*/ 				
		public function set context3D(value:Context3D):void {_context3D = value;}
		//----------------------------------------------------------------------------------------------	
	}

}
